@@ -1,11 +1,11 @@
 from argparse import ArgumentParser
 from collections import OrderedDict
 from copy import deepcopy
-from itertools import cycle
 import json
 from pprint import pformat
 import sys
 import time
+import warnings
 from typing import Any, Callable, Dict, List, Optional, Sequence, Set, Tuple, TypeVar, Union, get_type_hints
 from typing_inspect import is_literal_type, get_args, get_origin, is_union_type
 
@@ -20,9 +20,14 @@ from tap.utils import (
     type_to_str,
     get_literals,
     boolean_type,
-    TupleTypeEnforcer
+    TupleTypeEnforcer,
+    PythonObjectEncoder,
+    as_python_object,
+    fix_py36_copy
 )
 
+
+# Constants
 EMPTY_TYPE = get_args(List)[0] if len(get_args(List)) > 0 else tuple()
 
 SUPPORTED_DEFAULT_BASE_TYPES = {str, int, float, bool}
@@ -143,6 +148,7 @@ class Tap(ArgumentParser):
                     kwargs['nargs'] = kwargs.get('nargs', '*')
                 # Handle Tuple type (with type args) by extracting types of Tuple elements and enforcing them
                 elif get_origin(var_type) in (Tuple, tuple) and len(get_args(var_type)) > 0:
+                    loop = False
                     types = get_args(var_type)
 
                     # Don't allow Tuple[()]
@@ -152,12 +158,13 @@ class Tap(ArgumentParser):
 
                     # Handle Tuple[type, ...]
                     if len(types) == 2 and types[1] == Ellipsis:
-                        types = cycle([types[0]])
+                        types = types[0:1]
+                        loop = True
                         kwargs['nargs'] = '*'
                     else:
                         kwargs['nargs'] = len(types)
 
-                    var_type = TupleTypeEnforcer(types=types)
+                    var_type = TupleTypeEnforcer(types=types, loop=loop)
                 # To identify an Optional type, check if it's a union of a None and something else
                 elif (
                     is_union_type(var_type)
@@ -178,15 +185,15 @@ class Tap(ArgumentParser):
                     # If List or Set type, set nargs
                     if var_type in SUPPORTED_DEFAULT_COLLECTION_TYPES:
                         kwargs['nargs'] = kwargs.get('nargs', '*')
-                    
+
                     # Extract boxed type for Optional, List, Set
                     arg_types = get_args(var_type)
-                    
+
                     # Set defaults type to str for Type and Type[()]
                     if len(arg_types) == 0 or arg_types[0] == EMPTY_TYPE:
                         var_type = str
                     else:
-                        var_type = arg_types[0]  
+                        var_type = arg_types[0]
 
                     # Handle the cases of Optional[bool], List[bool], Set[bool]
                     if var_type == bool:
@@ -217,7 +224,7 @@ class Tap(ArgumentParser):
                 name_or_flags, kwargs = self.argument_buffer[variable]
                 self._add_argument(*name_or_flags, **kwargs)
             else:
-                flag_name = variable.replace("_", "-") if self._underscores_to_dashes else variable
+                flag_name = variable.replace('_', '-') if self._underscores_to_dashes else variable
                 self._add_argument(f'--{flag_name}')
 
         # Add any arguments that were added manually in add_arguments but aren't class variables (in order)
@@ -384,25 +391,101 @@ class Tap(ArgumentParser):
 
     def _get_argument_names(self) -> Set[str]:
         """Returns a list of variable names corresponding to the arguments."""
-        return set(self._get_class_dict().keys()) | set(self._annotations.keys())
+        return (set(self._get_class_dict().keys()) |
+                set(self._annotations.keys()) |
+                set(self.argument_buffer.keys()) - {'help'}) 
 
     def as_dict(self) -> Dict[str, Any]:
-        """Returns the member variables corresponding to the class variable arguments.
+        """Returns the member variables corresponding to the parsed arguments.
 
-         :return: A dictionary mapping each argument's name to its value.
-         """
+        Note: This does not include attributes set directly on an instance 
+        (e.g. arg is not included in MyTap().arg = "hi")
+
+        :return: A dictionary mapping each argument's name to its value.
+        """
         if not self._parsed:
             raise ValueError('You should call `parse_args` before retrieving arguments.')
 
         return {var: getattr(self, var) for var in self._get_argument_names()}
 
+    def from_dict(self, args_dict: Dict[str, Any], skip_unsettable: bool = False) -> None:
+        """Loads arguments from a dictionary, ensuring all required arguments are set.
+
+        :param args_dict: A dictionary from argument names to the values of the arguments.
+        :param skip_unsettable: When True, skips attributes that cannot be set in the Tap object,
+                                e.g. properties without setters.
+        """
+        # All of the required arguments must be provided or already set
+        required_args = {a.dest for a in self._actions if a.required}
+        unprovided_required_args = required_args - args_dict.keys()
+
+        for arg in unprovided_required_args:
+            if not hasattr(self, arg):
+                raise ValueError(f'Input dictionary "{args_dict}" does not include '
+                                 f'all unset required argument "{arg}".')
+
+        # Load all arguments
+        for key, value in args_dict.items():
+            try:
+                setattr(self, key, value)
+            except AttributeError:
+                if not skip_unsettable:
+                    raise AttributeError(f'Cannot set attribute "{key}" to "{value}". '
+                                         f'To skip arguments that cannot be set \n'
+                                         f'\t"skip_unsettable = True"')
+
+        self._parsed = True
+
     def save(self, path: str) -> None:
-        """Saves the arguments and reproducibility information in JSON format.
+        """Saves the arguments and reproducibility information in JSON format, pickling what can't be encoded.
 
         :param path: Path to the JSON file where the arguments will be saved.
         """
         with open(path, 'w') as f:
-            json.dump(self._log_all(), f, indent=4, sort_keys=True)
+            json.dump(self._log_all(), f, indent=4, sort_keys=True, cls=PythonObjectEncoder)
+
+    def load(self, path: str, check_reproducibility: bool = False) -> None:
+        """Loads the arguments in JSON format. Note: Due to JSON, tuples are loaded as lists.
+
+        :param path: Path to the JSON file where the arguments will be loaded from.
+        :param check_reproducibility: When True, raises an error if the loaded reproducibility
+                                      information doesn't match the current reproducibility information.
+        """
+        with open(path) as f:
+            args_dict = json.load(f, object_hook=as_python_object)
+
+        # Remove loaded reproducibility information since it is no longer valid
+        reproducibility = args_dict.pop('reproducibility', None)
+        
+        if check_reproducibility:
+            no_reproducibility_message = 'Reproducibility not guaranteed'
+            current_reproducibility = self.get_reproducibility_info()
+
+            if reproducibility is None:
+                raise ValueError(f'{no_reproducibility_message}: Could not find reproducibility '
+                                 f'information in args loaded from "{path}".')
+
+            if 'git_url' not in reproducibility:
+                raise ValueError(f'{no_reproducibility_message}: Could not find '
+                                 f'git url in args loaded from "{path}".')
+
+            if 'git_url' not in current_reproducibility:
+                raise ValueError(f'{no_reproducibility_message}: Could not find '
+                                 f'git url in current args.')
+
+            if reproducibility['git_url'] != current_reproducibility['git_url']:
+                raise ValueError(f'{no_reproducibility_message}: Differing git url/hash '
+                                 f'between current args and args loaded from "{path}".')
+
+            if reproducibility['git_has_uncommitted_changes']:
+                raise ValueError(f'{no_reproducibility_message}: Uncommitted changes '
+                                 f'in args loaded from "{path}".')
+
+            if current_reproducibility['git_has_uncommitted_changes']:
+                raise ValueError(f'{no_reproducibility_message}: Uncommitted changes '
+                                 f'in current args.')
+
+        self.from_dict(args_dict)
 
     def __str__(self) -> str:
         """Returns a string representation of self.
@@ -410,3 +493,18 @@ class Tap(ArgumentParser):
         :return: A formatted string representation of the dictionary of all arguments.
         """
         return pformat(self.as_dict())
+
+    @fix_py36_copy
+    def __deepcopy__(self, memo: Dict[int, Any] = None) -> TapType:
+        """Deepcopy the Tap object."""
+        copied = type(self).__new__(type(self))
+
+        if memo is None:
+            memo = {}
+
+        memo[id(self)] = copied
+
+        for (k, v) in self.__dict__.items():
+            copied.__dict__[k] = deepcopy(v, memo)
+
+        return copied
