@@ -1,12 +1,18 @@
 from argparse import ArgumentParser, ArgumentTypeError
+from base64 import b64encode, b64decode
 from collections import OrderedDict
+import copy
+from functools import wraps
 import inspect
 from io import StringIO
+from json import JSONEncoder
 import os
+import pickle
 import re
 import subprocess
+import sys
 import tokenize
-from typing import Any, Callable, Dict, Generator, Iterable, List, Tuple, Union
+from typing import Any, Callable, Dict, Generator, Iterator, List, Tuple, Union
 from typing_extensions import Literal
 from typing_inspect import get_args
 
@@ -230,14 +236,127 @@ def boolean_type(flag_value: str) -> bool:
 
 class TupleTypeEnforcer:
     """The type argument to argparse for checking and applying types to Tuples."""
-    def __init__(self, types: Iterable[type]):
-        self.types = (boolean_type if t == bool else t for t in types)
-        # self.index = 0
+    def __init__(self, types: List[type], loop: bool = False):
+        self.types = [boolean_type if t == bool else t for t in types]
+        self.loop = loop
+        self.index = 0
 
     def __call__(self, arg: str) -> Any:
-        return next(self.types)(arg)
+        arg = self.types[self.index](arg)
+        self.index += 1
 
-        # output = self.types[self.index](arg)
-        # self.index += 1
-        #
-        # return output
+        if self.loop:
+            self.index %= len(self.types)
+
+        return arg
+
+
+class MockTuple:
+    """Mock of a tuple needed to prevent JSON encoding tuples as lists."""
+    def __init__(self, _tuple: tuple) -> None:
+        self.tuple = _tuple
+
+
+def _nested_replace_type(obj: Any, find_type: type, replace_type: type) -> Any:
+    """Replaces any instance (including instances within lists, tuple, dict) of find_type with an instance of replace_type.
+
+    Note: Tuples, lists, and dicts are NOT modified in place.
+    Note: Does NOT do a nested search through objects besides tuples, lists, and dicts (e.g. sets).
+
+    :param obj: The object to modify by replacing find_type instances with replace_type instances.
+    :param find_type: The type to find in obj.
+    :param replace_type: The type to used to replace find_type in obj.
+    :return: A version of obj with all instances of find_type replaced by replace_type
+    """
+    if isinstance(obj, tuple):
+        obj = tuple(_nested_replace_type(item, find_type, replace_type) for item in obj)
+
+    elif isinstance(obj, list):
+        obj = [_nested_replace_type(item, find_type, replace_type) for item in obj]
+
+    elif isinstance(obj, dict):
+        obj = {
+            _nested_replace_type(key, find_type, replace_type): _nested_replace_type(value, find_type, replace_type)
+            for key, value in obj.items()
+        }
+
+    if isinstance(obj, find_type):
+        obj = replace_type(obj)
+
+    return obj
+
+
+class PythonObjectEncoder(JSONEncoder):
+    """Stores parameters that are not JSON serializable as pickle dumps.
+
+    See: https://stackoverflow.com/a/36252257
+    """
+    def iterencode(self, o: Any, _one_shot: bool = False) -> Iterator[str]:
+        o = _nested_replace_type(o, tuple, MockTuple)
+        return super(PythonObjectEncoder, self).iterencode(o, _one_shot)
+
+    def default(self, obj: Any) -> Any:
+        if isinstance(obj, set):
+            return {
+                '_type': 'set',
+                '_value': list(obj)
+            }
+        elif isinstance(obj, MockTuple):
+            return {
+                '_type': 'tuple',
+                '_value': list(obj.tuple)
+            }
+        return {
+            '_type': f'python_object (type = {obj.__class__.__name__})',
+            '_value': b64encode(pickle.dumps(obj)).decode('utf-8')
+        }
+
+
+def as_python_object(dct: Any) -> Any:
+    """The hooks that allow a parameter that is not JSON serializable to be loaded.
+
+    See: https://stackoverflow.com/a/36252257
+    """
+    if '_type' in dct and '_value' in dct:
+        _type, value = dct['_type'], dct['_value']
+
+        if _type == 'tuple':
+            return tuple(value)
+
+        elif _type == 'set':
+            return set(value)
+
+        elif _type.startswith('python_object'):
+            return pickle.loads(b64decode(value.encode('utf-8')))
+
+        else:
+            raise ValueError(f'Special type "{_type}" not supported for JSON loading.')
+
+    return dct
+
+
+def fix_py36_copy(func: Callable) -> Callable:
+    """Decorator that fixes functions using Python 3.6 deepcopy of ArgumentParsers.
+
+    Based on https://stackoverflow.com/questions/6279305/typeerror-cannot-deepcopy-this-pattern-object
+    """
+    if sys.version_info[:2] > (3, 6):
+        return func
+
+    @wraps(func)
+    def wrapper(*args, **kwargs):
+        re_type = type(re.compile(''))
+        has_prev_val = re_type in copy._deepcopy_dispatch
+        prev_val = copy._deepcopy_dispatch.get(re_type, None)
+        copy._deepcopy_dispatch[type(re.compile(''))] = lambda r, _: r
+
+        result = func(*args, **kwargs)
+
+        if has_prev_val:
+            copy._deepcopy_dispatch[re_type] = prev_val
+        else:
+            del copy._deepcopy_dispatch[re_type]
+
+        return result
+
+    return wrapper
